"""Utilities for generating hyperparameter plans."""

from typing import Any, Dict, Iterator, Iterable, Mapping, Optional, Union
from dataclasses import dataclass
import pickle
from io import TextIOWrapper
from itertools import product
import json

from carabiner import cast, print_err
import numpy as np


@dataclass
class HyperRange:

    """Produce sequences evenly spaced in log or linear space.

    Simply a convenience wrapper around `np.linspace` and `np.geomspace`
    to accept boolean types, ensure types are maintained, and behave 
    like builtin `range`.

    Parameters
    ----------
    min : float | int | bool
        Minimum value when `max` provided, else the maximum value of the range. 
    max : float | int | bool, optional
        Maximum value of the range if minimum is provided.
    n_step : int
        Number of steps in the sequence. Default: 2 if 
        `min` and `max` are different, else 1.
    log_scale : bool
        Whether to have log spacing in the sequence. Default: `False`.

    Returns
    -------
    HyperRange

    Examples
    --------
    >>> HyperRange(1.)
    HyperRange: [ 1.0 ]
    >>> HyperRange(1., 10.)
    HyperRange: [ 1.0, 10.0 ]
    >>> HyperRange(1., 100., 20)
    HyperRange: [ 1.0, 6.2105263157894735, ..., 94.78947368421052, 100.0 ]
    >>> HyperRange(False, True)
    HyperRange: [ False, True ]
    >>> HyperRange(False, True) == HyperRange(True, False)
    True
    >>> HyperRange(False, True).min
    False
    >>> HyperRange(False, True).max
    True
    >>> HyperRange(1, 100).max
    100
    >>> HyperRange(1, 100).min
    1
    >>> HyperRange(1, 100.1).max  # Output type is the same as `start`
    100
    >>> list(HyperRange(1, 100))
    [1, 100]
    
    """
    
    start: Union[float, int, bool]
    stop: Optional[Union[float, int, bool]] = None
    n_step: Optional[int] = None
    log_scale: bool = False

    def __post_init__(self):
        range_fn = np.linspace if not self.log_scale else np.geomspace
        if self.stop is None:
            self.stop = self.start
        if self.n_step is None:
            if self.stop == self.start:
                self.n_step = 1
            else:
                 self.n_step = 2
        self.range = sorted(set(
            range_fn(
                self.start, 
                self.stop, 
                num=self.n_step,
            )
            .astype(type(self.start))
            .tolist()
        ))

    def __iter__(self) -> Iterator[Union[float, int, bool]]:
        yield from self.range

    def __len__(self) -> int:
        return len(self.range)

    def __eq__(self, b) -> bool:
        if isinstance(b, type(self)):
            return self.range == b.range
        else:
            return self.range == b

    def __repr__(self) -> str:
        prefix = "HyperRange"

        if len(self) > 4:
            lead = ', '.join(map(str, self.range[:2]))
            trail = ', '.join(map(str, self.range[-2:]))
            return f"{prefix}: [ {lead}, ..., {trail} ]"
        else:
            return f"{prefix}: [ {', '.join(map(str, self.range))} ]"

    def __str__(self) -> str:
        return str(self)
        
    @property
    def min(self) -> Union[float, int, bool]:
        """Get minimum value.

        Returns
        -------
        float | int | bool
            Minimum value of HyperRange.

        """
        return min(self.range)

    @property
    def max(self) -> Union[float, int, bool]:
        """Get maximum value.

        Returns
        -------
        float | int | bool
            Maximum value of HyperRange.

        """
        return max(self.range)
          

class HyperOpt(dict):

    """Construct a parameter grid based on the provided ranges.

    Examples
    --------
    >>> HyperOpt(a=1.)
    {'a': HyperRange: [ 1.0 ]}
    >>> HyperOpt(a=1., b=True)
    {'a': HyperRange: [ 1.0 ], 'b': HyperRange: [ True ]}
    >>> configs = HyperOpt(a=1., b=True, d=HyperRange(50, 53, 4))
    >>> len(configs)
    4
    >>> configs
    {'a': HyperRange: [ 1.0 ], 'b': HyperRange: [ True ], 'd': HyperRange: [ 50, 51, 52, 53 ]}
    >>> for conf in configs: print(conf)
    ... 
    {'a': 1.0, 'b': True, 'd': 50}
    {'a': 1.0, 'b': True, 'd': 51}
    {'a': 1.0, 'b': True, 'd': 52}
    {'a': 1.0, 'b': True, 'd': 53}

    """

    def __init__(
        self, 
        silent: bool = False, 
        _ranges: Optional[Iterable[Mapping]] = None, 
        *args, **kwargs
    ):
        super().__init__(self, *args, **kwargs)
        for key in self.keys():
            val = self[key]
            if not isinstance(val, HyperRange):
                val = cast(val, to=list)
                if isinstance(val[0], (float, int, bool)) and len(val) <= 2:
                    self[key] = HyperRange(*val)
                else:
                    self[key] = val
        iterator = product(*(self.get(key) for key in self.keys()))
        configs = tuple(dict(zip(self.keys(), args)) for args in iterator)
        if not silent:
            print_err(f"There are {len(configs)} configurations to test.")
        self._ranges = tuple(_ranges or []) + configs
        self._keys = tuple(self._ranges[0])

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        yield from self._ranges

    def __len__(self) -> int:
        return len(self._ranges)

    @classmethod
    def from_file(
        cls, 
        file: Union[str, TextIOWrapper],
        serialized: bool = False,
        **kwargs
    ):
        """Create a HyperOpt object from a JSON or serialized file.

        """
        if not serialized:
            f = cast(file, to=TextIOWrapper)
            d = json.load(f)
            return cls(**kwargs, **d)
        else:
            with open(file, "rb") as f:
                obj = pickle.load(f)
            return obj
            
    def write(
        self, 
        file: Union[str, TextIOWrapper],
        as_list: bool = False,
        serialize: bool = False
    ) -> None:
        """Save a HyperOpt object in a reloadable format.

        """
        file = cast(file, str)
        open_mode = "wb" if serialize else "w"
        with open(file, open_mode) as f:
            if serialize:
                pickle.dump(obj=self, file=f)
            else:
                data = {key: list(val) for key, val in self.items()}
                if as_list:
                    data.update({"_ranges": self._ranges})
                json.dump(
                    data, 
                    f,
                    sort_keys=False,  # ensure round trip yields same order
                    indent=4,
                )
        return None



"""Plotting utilities."""

from carabiner.mpl import add_legend, grid, figsaver
import pandas as pd
import numpy as np


def _plot_history(
    lightning_csv, 
    filename: str
) -> None:

    data_to_plot = (
        pd.read_csv(lightning_csv)
        .groupby(['epoch', 'step'])
        .agg(np.nanmean)
        .reset_index()
    )

    fig, ax = grid(aspect_ratio=1.5)
    for _y in ('val_loss', 'loss', 'learning_rate'):
        if _y in data_to_plot:
            ax.plot(
                'step', _y, 
                data=data_to_plot, 
                label=_y,
            )
            ax.scatter(
                'step', _y, 
                data=data_to_plot,
                s=1.,
            )
    add_legend(ax)
    ax.set(
        xlabel='Training step', 
        ylabel='Loss', 
        yscale='log',
    )
    figsaver(format="png")(fig, name=filename, df=data_to_plot)
    return None


def _plot_prediction_scatter(
    df,
    filename: str,
    x: str = "__prediction__",
    y: str = "labels"
) -> None:
    fig, ax = grid()
    ax.scatter(
        x, y,
        data=df,
        s=1.,
    )
    ax.plot(
        ax.get_ylim(),
        ax.get_ylim(),
        color='dimgrey',
        zorder=-5,
    )
    ax.set(
        xlabel=f"Predicted ({x})", 
        ylabel=f"Observed ({y})",
    )
    figsaver(format="png")(
        fig,
        filename,
        df=df,
    )
    return None



"""Utilities used in splitting."""

from typing import Dict, Mapping, Union

from carabiner import print_err
from datasets import Dataset, IterableDataset
from tqdm.auto import tqdm


def annotate_split(
    x: Mapping[str, str], 
    key_to_split: Mapping[str, str],
    key_column: str,
    split_column: str = "split"
) -> Dict[str, ...]:
    """Tag each entry with its split label based on union-find component.

    Examples
    ========
    >>> import numpy as np
    >>> comp_map = {"A": "train", "B": "test"}
    >>> out = annotate_split({"col": ["B", "A"]}, key_to_split=comp_map, key_column="col")
    >>> out['split']
    ['test', 'train']

    """
    x[split_column] = [key_to_split.get(k) for k in x[key_column]]
    return x


def dataset_len(ds: Union[Dataset, IterableDataset]):
    if isinstance(ds, Dataset):
        return ds.num_rows
    elif isinstance(ds, IterableDataset):
        print_err("Looping through dataset to count rows: ", end="")
        for i, _ in enumerate(tqdm(ds)):
            pass
        return i
    else:
        raise ValueError(f"Input must be Dataset or IterableDataset but was `{type(ds)}`.")


"""Functions in pytorch."""

import torch 

from ..stateless.typing import ArrayLike, Array


def mse_loss(
    y_pred: ArrayLike, 
    y_true: ArrayLike
) -> Array:
    return torch.mean(torch.square(y_pred - y_true))



"""Initializing ModelBoxes automatically."""

from typing import Optional

from carabiner import print_err

from .checkpoint_utils import load_checkpoint_file
try:
    from .torch.modelbox import *
except ImportError as e:
    print_err(e)
    raise ImportError(
        """
        Modelling is not installed. Try reinstalling duvida with:

            $ pip install duvida[torch]

        or to use chemistry ML/AI:

            $ pip install duvida[chem]

        """
    )
else:
    from .base.modelboxes import ModelBoxBase
    from .base.modelbox_registry import DEFAULT_MODELBOX, MODELBOX_REGISTRY

class AutoModelBox:

    _init_kwargs_file: str = ModelBoxBase._init_kwargs_filename
    _model_config_file: str = ModelBoxBase._model_config_filename
    _model_class_key: str = "class_name"

    def __init__(
        self,
        class_name: str = DEFAULT_MODELBOX,
        **kwargs
    ):  
        self._class = MODELBOX_REGISTRY[class_name]
        self._instance = self._class(**{
            key: val for key, val in kwargs.items() 
            if key != self._model_class_key
        })

    @classmethod
    def from_pretrained(
        cls,
        checkpoint: str,
        cache_dir: Optional[str] = None,
        **kwargs
    ) -> ModelBoxBase:
        config_file = cls._init_kwargs_file
        config = load_checkpoint_file(
            checkpoint=checkpoint,
            filename=config_file,
            callback="json",
            cache_dir=cache_dir,
            **kwargs
        )
        modelbox = cls(**config)._instance
        modelbox.load_checkpoint(checkpoint, cache_dir=cache_dir)
        modelbox._default_cache = cache_dir
        return modelbox

    @classmethod
    def show(cls):
        return tuple(MODELBOX_REGISTRY)



import numpy as np
import pytest
from duvida.base.evaluation import rmse, pearson_r, spearman_r


def test_rmse_simple():
    pred = [1., 2., 3.]
    truth = [1., 2., 4.]
    assert rmse(pred, truth) == pytest.approx(np.sqrt(((np.asarray(pred) - np.asarray(truth)) ** 2).mean()))


def test_rmse_ensemble():
    pred = [[1., 2.], [2., 3.], [3., 4.]]
    truth = [1., 2., 4.]
    expected = np.sqrt(((np.asarray(pred).mean(axis=-1) - np.asarray(truth)) ** 2).mean())
    assert rmse(pred, truth) == pytest.approx(expected)


def test_rmse_broadcast():
    pred = [1., 2., 3.]
    truth = [[1.], [2.], [4.]]
    expected = np.sqrt(((np.asarray(pred) - np.asarray(truth)) ** 2).mean())
    assert rmse(pred, truth) == pytest.approx(expected)


def test_rmse_fail_shape():
    pred = np.ones((2, 2))
    truth = np.ones(3)
    with pytest.raises(ValueError):
        rmse(pred, truth)


def test_pearson_and_spearman():
    x = [[1., 2.], [2., 4.]]
    y = [1., 3.]
    r = pearson_r(x, y)
    rho = spearman_r(x, y)
    assert r == pytest.approx(1.)
    assert rho == pytest.approx(1.)



"""Utilities for evaluating predictions."""

from typing import Callable, Tuple

import numpy as np
from numpy.typing import ArrayLike
import scipy

def _normalize_dims(
    x: ArrayLike, 
    y: ArrayLike,
    broadcast: bool = False,
    aggfun: Callable = np.mean
) -> Tuple[np.ndarray, np.ndarray]:
    """Normalize prediction dimensions.

    If `x` has one extra dimension compared to `y` and their leading
    dimensions match exactly, `x` is assumed to be an ensemble of
    predictions and the mean across its last axis is used. Otherwise the
    function ensures that `x` and `y` are broadcastable. and raises a
    ``ValueError`` if they are not.

    Parameters
    ==========
    x : ArrayLike
        Prediction array.
    y : ArrayLike
        Ground truth array.

    Returns
    =======
    Tuple[numpy.ndarray, numpy.ndarray]
        Possibly averaged `x` and original `y`.

    Raises
    ======
    ValueError
        If `x` and `y` are not broadcastable.
    
    """

    x, y = np.asarray(x), np.asarray(y)

    if x.shape[:-1] != y.shape[:(x.ndim - 1)]:
        raise ValueError(
            f"Leading dimensions of x and y (shapes {x.shape} and {y.shape}) are not compatible!"
        )   

    if x.ndim == y.ndim + 1:
        x = aggfun(x, axis=-1)
    elif x.ndim == y.ndim and y.shape[-1] == 1:
        x = aggfun(x, axis=-1, keepdims=True)

    try:
        b = np.broadcast(x, y)
    except ValueError:
        raise ValueError(
            f"Shapes {x.shape} and {y.shape} are not broadcastable!"
        )

    if broadcast:
        bshape = b.shape
        x, y = (np.broadcast_to(v, bshape).ravel() for v in (x, y))

    return x, y


def rmse(x, y):
    """Root mean square error.

    Parameters
    ----------
    x : numpy.ndarray
        Predicted values. If predictions are provided as an ensemble, the mean
        across the last axis is used.
    y : numpy.ndarray
        Ground truth values.

    Returns
    -------
    float
        The RMSE between ``x`` and ``y``.
    
    Examples
    --------
    >>> import numpy as np
    >>> pred = np.array([1.0, 2.0, 3.0])
    >>> truth = np.array([1.0, 2.0, 4.0])
    >>> np.allclose(rmse(pred, truth), np.sqrt(((pred - truth) ** 2).mean()))
    True

    """

    x, y = _normalize_dims(x, y)
    return np.sqrt(np.mean(np.square(x - y)))


def mae(x, y):
    x, y = _normalize_dims(x, y)
    return np.mean(np.abs(x - y))


def pearson_r(x, y):
    x, y = _normalize_dims(x, y, broadcast=True)
    return  np.corrcoef(x, y)[0, 1]


def spearman_r(x, y):
    x, y = _normalize_dims(x, y, broadcast=True)
    return scipy.stats.spearmanr(
        x, 
        y, 
        nan_policy="omit",
    ).statistic



"""Data preprocessing functions."""

from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Union
from functools import partial

from chemprop.data import (
    MoleculeDatapoint, 
    MoleculeDataset, 
    MolGraph
)
import numpy as np
from schemist.features import calculate_feature

from .registry import register_function


@register_function("identity")
def Identity() -> Callable:
    """Simple pass-through.
    
    """
    def _identity(
        data: Mapping[str, Iterable],
        input_column: str
    ) -> np.ndarray:
        return np.asarray(data[input_column])
    return _identity


@register_function("one-hot")
def OneHot(
    categories: Iterable[str],
    intercept: bool = False
) -> Callable:
    """Convert string labels into one-hot encodings.
    
    """
    prepend = [1] if intercept else []

    def _one_hot(
        data: Mapping[str, Iterable],
        input_column: str
    ) -> np.ndarray:
        return np.asarray([
            prepend + [1 if x == cat else 0 for cat in categories] 
            for x in data[input_column]
        ])

    return _one_hot
    

@register_function("morgan-fingerprint")
def MorganFingerprint(**kwargs) -> Callable:
    """Get Morgan fingerprint from SMILES.
    
    """
    feature_calculator = partial(
        calculate_feature,
        feature_type="fp",
        return_dataframe=False,
        on_bits=False,
        **kwargs,
    )

    def _morgan_fingerprint(
        data: Mapping[str, Iterable],
        input_column: str
    ) -> np.ndarray:
        fingerprints, _ = feature_calculator(strings=data[input_column])
        return fingerprints
        
    return _morgan_fingerprint


@register_function("descriptors-2d")
def Descriptors2D(
    normalized: bool = True,
    histogram_normalized: bool = True
) -> Callable:
    """Get 2D descriptors from SMILES, optionally normalized.
    
    """
    feature_calculator = partial(
        calculate_feature,
        feature_type="2d",
        return_dataframe=False,
        normalized=normalized,
        histogram_normalized=histogram_normalized,
    )
    
    def _descriptors_2d(
        data: Mapping[str, Iterable],
        input_column: str
    ) -> np.ndarray:
        desc_2d, _ = feature_calculator(strings=data[input_column])
        return desc_2d

    return _descriptors_2d


@register_function("chemprop-mol")
def ChempropData(
    label_column: Optional[Union[str, Iterable[str]]] = None,
    extra_featurizers: Optional[Union[Mapping[str, Any], Iterable[Mapping[str, Any]]]] = None
) -> Callable:
    """Convert SMILES to iterable of Chemprop datum.
    
    """
    if isinstance(extra_featurizers, str):
        extra_featurizers = [extra_featurizers]
    if isinstance(label_column, str):
        label_column = [label_column]

    def _stack_columns(
        data: Mapping[str, Iterable],
        nrows: int,
        columns: Optional[Iterable[str]] = None
    ) -> np.ndarray:
        if columns is None:
            array = [None] * nrows
        else: 
            array = [np.asarray(data[col]) for col in columns]
            array = [a if a.ndim > 1 else a[..., np.newaxis] for a in array]
            if len(array) > 0:
                array = np.concatenate(array, axis=-1).astype(np.float32) 
            else:
                array = [None] * nrows
        return array
    
    def _chemprop_data(
        data: Mapping[str, Iterable],
        input_column: str
    ) -> List[Dict[str, np.ndarray]]:
        nrows = len(data[input_column])
        y_vals = _stack_columns(data, nrows, label_column)
        extra_features = _stack_columns(data, nrows, extra_featurizers)

        mol_datapoints = [
            MoleculeDatapoint.from_smi(smi=x, y=y, x_d=xd) 
            for x, y, xd in zip(data[input_column], y_vals, extra_features)
        ]
    
        datums = []
        for datum in MoleculeDataset(mol_datapoints):
            new_datum = {}
            for key, val in datum._asdict().items():
                if isinstance(val, MolGraph):
                    new_val = {
                        key2: val2.astype(np.float32) if isinstance(val2, np.ndarray) else np.float32(val2) 
                        for key2, val2 in val._asdict().items()
                    }
                elif isinstance(val, float):
                    new_val = np.float32(val)
                elif isinstance(val, np.ndarray):
                    new_val = val.astype(np.float32)
                elif val is not None:
                    new_val = val
                else:
                    new_val = None
                new_datum[key] = new_val
            datums.append(new_datum)
        return datums

    return _chemprop_data



"""Modelbox data aspects."""

from typing import Callable, Dict, Mapping, Union
from functools import partial
from multiprocessing import cpu_count

from datasets import Dataset, IterableDataset
import torch
from torch.utils.data import DataLoader

from ...base.data import ChemMixinBase, DataMixinBase
from ...stateless.config import config

config.set_backend('torch', precision='float')

from ...stateless.typing import Array, ArrayLike
from ...stateless.utils import vmap
from ..models.chemprop.data import _collate_training_batch_for_forward

_DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class DataMixin(DataMixinBase):

    _format = 'pytorch'
    _format_kwargs = {
        'dtype': torch.float32, 
        'device': _DEVICE,
    }

    @staticmethod
    def make_dataloader(
        dataset: Union[Dataset, IterableDataset],
        batch_size: int = 16, 
        shuffle: bool = False,
        **kwargs
    ) -> DataLoader:
        # kwargs = {"num_workers": min(4, cpu_count() - 1)} | kwargs
        # kwargs = {"num_workers": 1} | kwargs
        return DataLoader(
            dataset, 
            batch_size=batch_size, 
            shuffle=shuffle,
            **kwargs,
        )


class TorchChemMixin(ChemMixinBase, DataMixin):

    @staticmethod
    def _get_max_sim(
        query: ArrayLike, 
        references: ArrayLike,
        aggregator: Callable[[ArrayLike], float] = torch.max
    ) -> Array:
        a_n_b = torch.sum(query.unsqueeze(0) * references, dim=-1, keepdim=True)
        similarities = a_n_b / (torch.sum(query) + torch.sum(references, dim=-1) - torch.sum(a_n_b, dim=-1)).unsqueeze(-1)
        return aggregator(similarities)

    @staticmethod
    def _get_nn_tanimoto(
        x: Mapping[str, ArrayLike],
        refs_data: Mapping[str, ArrayLike],
        results_column: str,
        _in_key: str,
        _sim_fn: Callable[[ArrayLike, ArrayLike], float]
    ) -> Dict[str, Array]:
        query_fps = x[_in_key]
        refs = refs_data[_in_key]
        results = vmap(_sim_fn, in_axes=(0, None))(query_fps, refs)
        x[results_column] = results
        return x


class ChempropDataMixin(TorchChemMixin):

    @staticmethod
    def make_dataloader(
        dataset: Dataset, 
        batch_size: int = 16, 
        shuffle: bool = False
    ) -> DataLoader:
        return DataLoader(
            dataset, 
            batch_size=batch_size, 
            shuffle=shuffle,
            collate_fn=partial(
                _collate_training_batch_for_forward, 
                for_dataloader=True,
            ),
        )



"""Base mixins for training."""

from abc import abstractmethod, ABC


class ModelTrainerBase(ABC):

    def __init__(
        self, 
        epochs: int, 
        *args, **kwargs,
    ):
        self.epochs = epochs
        self._args = args
        self._kwargs = kwargs
        self._trainer = None

    @abstractmethod
    def create_trainer(self):
        pass

    @abstractmethod
    def train(self, model, train_dataloader, val_dataloader):
        pass



"""Wrapping core algorithms with convenience."""

from typing import Callable, Mapping, Optional

from functools import wraps

def process_splits(f: Callable):

    @wraps(f)
    def _process_splits(
        *args,
        train: float = 1.,
        validation: Optional[float] = None,
        test: Optional[float] = None,
        splits: Optional[Mapping[str, float]] = None,
        **kwargs
    ) -> Callable:
        validation = validation or (1. - train)
        test = test or (1. - train - validation)
        kwarg_splits = {
            "train": train,
        }
        if validation > 0.:
            kwarg_splits["validation"] = validation
        if test > 0.:
            kwarg_splits["test"] = test
        splits = splits or {}
        kwarg_splits.update(splits)  # override defaults, add arbitrary extra

        return f(*args, **kwargs, splits=kwarg_splits)

    return _process_splits



"""Algorithm for packing groups into splits."""

from typing import Dict, Mapping, Optional, Union

from collections import Counter

from carabiner import print_err
from datasets import Dataset, IterableDataset
import numpy as np
from tqdm.auto import tqdm

from .utils import dataset_len

def deterministic_packing_step(
    group_size: int,
    group_name: str,
    remaining: Mapping[str, int],
    group_to_split: Mapping[str, str],
    **kwargs
):
    split_destination = max(remaining, key=remaining.__getitem__)
    group_to_split[group_name] = split_destination
    remaining[split_destination] -= group_size
    return group_to_split, remaining


def random_packing_step(
    group_size: int,
    group_name: str,
    remaining: Mapping[str, int],
    group_to_split: Mapping[str, str],
    rng
):
    eligible = [
        split_name for split_name, capacity in remaining.items() 
        if capacity >= group_size
    ]
    if len(eligible) == 0:
        # if none can fully fit, fall back to all splits (you may accept overflow)
        eligible = list(remaining)

    weights = np.array(
        [remaining[name] for name in eligible], 
        dtype=float,
    )
    probs = weights / weights.sum()

    split_destination = rng.choice(eligible, p=probs)
    group_to_split[group_name] = split_destination
    remaining[split_destination] -= group_size
    return group_to_split, remaining

def pack_bins(
    ds: Union[Dataset, IterableDataset],
    group_column: str,
    splits: Mapping[str, float],
    batch_size: int = 1024,
    num_rows: Optional[int] = None,
    deterministic: bool = True,
    seed: int = 42
) -> Dict[str, str]:
    rng = np.random.default_rng(seed=seed)
    if num_rows is None:
        num_rows = dataset_len(ds)

    sizes = Counter()
    for example in tqdm(ds.iter(batch_size=batch_size), desc="Finding unique groups"):
        sizes.update(example[group_column])

    print_err(f"There are {len(sizes)} unique groups.")
    
    group_to_split = {}
    split_target_sizes = {
        key: np.ceil(num_rows * val) for key, val in splits.items()
    }
    if deterministic:
        packing_f = deterministic_packing_step
    else:
        packing_f = random_packing_step
    for group in tqdm(sorted(
        sizes, 
        key=sizes.__getitem__, 
        reverse=True,
    ), desc="Sorting groups by size"):
        group_to_split, split_target_sizes = packing_f(
            group_size=sizes[group],
            group_name=group,
            remaining=split_target_sizes,
            group_to_split=group_to_split,
            rng=rng,
        )
            
    return group_to_split



"""Generating splits on chemical scaffold."""

from typing import Iterable, Optional, Mapping, Union

from datasets import Dataset, IterableDataset
from schemist.converting import convert_string_representation

from .bin_packing import pack_bins
from .decorators import process_splits
from .utils import annotate_split


def annotate_scaffold(
    x: Mapping[str, Iterable], 
    structure_column: str, 
    input_representation: str = 'smiles',
    scaffold_classifier: str = "scaffold",
    scaffold_column: str = "scaffold"
):
    """Annotate a chemical structure with its scaffold.

    """

    converted = convert_string_representation(
        strings=x[structure_column],
        input_representation=input_representation,
        output_representation=scaffold_classifier,
    )
    if len(x[structure_column]) > 1 and not isinstance(x[structure_column], str):
        x[scaffold_column] = list(converted)
    else:
        x[scaffold_column] = [converted]
    return x


@process_splits
def scaffold_split(
    ds: Union[Dataset, IterableDataset],
    structure_column: str,
    input_representation: str = 'smiles',
    scaffold_classifier: str = "scaffold",
    scaffold_column: str = "scaffold",
    deterministic: bool = True,
    seed: int = 42,
    batch_size: int = 1024,
    splits: Optional[Mapping[str, float]] = None
) -> Union[Dataset, IterableDataset]:

    """Spilt a dataset based on chemical scaffold (or other grouping calculation).

    """

    ds = (
        ds.
        map(
            annotate_scaffold,
            fn_kwargs={
                "structure_column": structure_column,
                "input_representation": input_representation,
                "scaffold_classifier": scaffold_classifier,
                "scaffold_column": scaffold_column,
            },
            batched=True,
            batch_size=batch_size,
            desc="Annotating scaffold",
        )
    )

    scaffold_to_split = pack_bins(
        ds,
        group_column=scaffold_column,
        splits=splits,
        batch_size=batch_size,
        deterministic=deterministic,
        seed=seed,
    )

    if isinstance(ds, Dataset):
        desc = {"desc": "Annotating splits"}
    else:
        desc = {}
    ds = ds.map(
        annotate_split,
        fn_kwargs={
            "key_to_split": scaffold_to_split,
            "key_column": scaffold_column,
            "split_column": "split",
        },
        batched=True,
        batch_size=batch_size,
        **desc,
    )

    if isinstance(ds, Dataset):
        desc = {"desc": "Generating split datasets"}
    else:
        desc = {}
    return ds, {
        key: ds.filter(lambda x: x["split"] == key, **desc)
        for key in splits
    }



"""Multi-layer perceptrons."""

from typing import Callable, List, Optional

from carabiner import cast, print_err
from torch.nn import BatchNorm1d, Dropout, Identity, Linear, Module, SiLU, Sequential

from torch.optim import Adam, Optimizer

from .utils.ensemble import TorchEnsembleMixin
from .utils.lt import LightningMixin
from ...stateless.config import config

config.set_backend('torch', precision='float')

from ...stateless.typing import Array, ArrayLike
from ...stateless.utils import jit

_DEFAULT_ACTIVATION: Callable[..., Module] = SiLU  # Smooth activation to prevent gradient collapse
_DEFAULT_N_UNITS: int = 16

class LinearStack(Module):

    @staticmethod
    def _add_layer(
        in_features: int,
        out_features: int,
        layers: Optional[list] = None,
        layer_class: Callable[..., Module] = Linear,
        batch_norm: bool = False,
        dropout: float = 0.0,
        activation: Callable[..., Module] = _DEFAULT_ACTIVATION,
        **kwargs
    ) -> List[Module]:
        layers = cast(
            [] if layers is None else layers, 
            to=list,
        )
        layers += [layer_class(in_features, out_features, **kwargs)]
        if batch_norm:
            layers.append(BatchNorm1d(out_features))
        layers.append(activation())
        if dropout > 0.:
            layers.append(Dropout(dropout))
        return layers

    def build_model(
        self, 
        n_input: int,
        n_out: int = 1,
        n_hidden: int = 1,
        n_units: int =_DEFAULT_N_UNITS,
        batch_norm: bool = False,
        dropout: float = 0.,
        activation: Callable[..., Module] = _DEFAULT_ACTIVATION,
        layer_class: Callable[..., Module] = Linear,
        **layer_kwargs
    ) -> Module:
        common_kwargs = {
            "out_features": n_units,
            "layer_class": layer_class,
            "batch_norm": batch_norm,
            "dropout": dropout,
            "activation": activation,
        }
        layers = self._add_layer(
            in_features=n_input,
            **common_kwargs,
            **layer_kwargs,
        )
        for _ in range(1, n_hidden):
            layers = self._add_layer(
                in_features=n_units,
                layers=layers,
                **common_kwargs,
                **layer_kwargs,
            )
        layers.append(
            Linear(in_features=n_units, out_features=n_out)
        )
        return Sequential(*layers)


class TorchResidualBlock(LinearStack):

    """Residual block module.

    Examples
    ========
    >>> import torch
    >>> net = TorchResidualBlock(n_input=4, residual_depth=2, n_units=8, n_out=1)
    >>> net(torch.randn(3, 4)).shape 
    torch.Size([3, 1])

    """

    def __init__(
        self, 
        n_input: int, 
        n_out: int, 
        residual_depth: int = 1,
        n_units: int =_DEFAULT_N_UNITS, 
        dropout: float = 0., 
        activation: Callable = _DEFAULT_ACTIVATION,  
        batch_norm: bool = False
    ):
        super().__init__()
        self.n_input = n_input
        self.residual_depth = residual_depth
        self.n_units = n_units
        self.dropout = dropout
        self.activation = activation
        self.n_out = n_out
        self.batch_norm = batch_norm
        if self.n_input == self.n_out:
            self.projection = Identity()
        else:
            self.projection = Linear(n_input, n_out)
        self.model_layers = self.build_model()

    def build_model(self):
        return super().build_model(
            n_input=self.n_input,
            n_out=self.n_out,
            n_hidden=self.residual_depth,
            n_units=self.n_units,
            batch_norm=self.batch_norm,
            dropout=self.dropout,
            activation=self.activation,
            layer_class=Linear,
        )

    # @jit  # Fails compile
    def forward(self, x: ArrayLike) -> Array:
        residual = self.model_layers(x)
        projection = self.projection(x)
        return self.activation()(residual + projection)


class TorchMLPBase(LinearStack):

    """Multilayer perceptron.

    Examples
    ========
    >>> import torch
    >>> net = TorchMLPBase(n_input=4, n_hidden=2, n_units=8, n_out=1)
    >>> net(torch.randn(3, 4)).shape 
    torch.Size([3, 1])
    >>> resnet = TorchMLPBase(n_input=4, n_hidden=4, n_units=8, n_out=1, residual_depth=2)
    >>> resnet(torch.randn(3, 4)).shape 
    torch.Size([3, 1])

    """

    def __init__(
        self, 
        n_input: int, 
        n_hidden: int = 1,
        n_units: int = _DEFAULT_N_UNITS, 
        dropout: float = 0., 
        activation: Callable = _DEFAULT_ACTIVATION,
        n_out: int = 1, 
        batch_norm: bool = False,
        residual_depth: Optional[int] = None
    ):
        super().__init__()
        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.dropout = dropout
        self.activation = activation
        self.n_out = n_out
        self.batch_norm = batch_norm
        if residual_depth is not None and (residual_depth > self.n_hidden):
            print_err(
                f"""
                WARNING: Skip length must be greater than number of hidden layers:
                - Skip length: {residual_depth}
                - Number of hidden layers: {self.n_hidden}
                Falling back to non-residual.
                """
            )
            self.residual_depth = None
        else:
            self.residual_depth = residual_depth
        if self.residual_depth is not None:
            self._layer_class = TorchResidualBlock
            self._layer_kwargs = {
                "residual_depth": self.residual_depth,
            }
            self._n_residual_blocks = self.n_hidden // (self.residual_depth or 1)
            self._n_extra_linear = self.n_hidden % (self.residual_depth or 1)
        else:
            self._layer_class = Linear
            self._layer_kwargs = {}
            self._n_residual_blocks = 0
            self._n_extra_linear = 0
        self.model_layers = self.build_model()

    def build_model(self):
        layers = super().build_model(
            n_input=self.n_input,
            n_out=(self.n_out if self._n_extra_linear == 0 else self.n_units), 
            layer_class=self._layer_class,
            n_hidden=(self._n_residual_blocks if self.residual_depth is not None else self.n_hidden),
            n_units=self.n_units, 
            dropout=self.dropout, 
            activation=self.activation,  
            batch_norm=self.batch_norm,
            **self._layer_kwargs,
        )
        if self._n_extra_linear > 0:
            extra_layers = super().build_model(
                n_input=self.n_units,
                n_out=self.n_out, 
                layer_class=Linear,
                n_hidden=self._n_extra_linear,
                n_units=self.n_units, 
                dropout=self.dropout, 
                activation=self.activation,  
                batch_norm=self.batch_norm,
            )
            layers = Sequential(layers, extra_layers)

        return layers
        
    @jit
    def forward(self, x: ArrayLike) -> Array:
        return self.model_layers(x)


class TorchMLPLightning(TorchMLPBase, LightningMixin):

    """Multilayer perceptron with Lightning.

    Examples
    ========
    >>> import torch
    >>> net = TorchMLPLightning(n_input=4, n_hidden=2, n_units=8, n_out=1)
    >>> net(torch.randn(3, 4)).shape 
    torch.Size([3, 1])

    """

    def __init__(
        self, 
        learning_rate: float = .01,
        optimizer: Optimizer = Adam,
        reduce_lr_on_plateau: bool = True,
        reduce_lr_patience: int = 10,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.save_hyperparameters()
        self._init_lightning(
            optimizer=optimizer, 
            learning_rate=learning_rate, 
            model_attr='model_layers',
            reduce_lr_on_plateau=reduce_lr_on_plateau,
            reduce_lr_patience=reduce_lr_patience,
        )


class TorchMLPEnsemble(TorchEnsembleMixin, TorchMLPLightning):

    """Multilayer perceptron ensemble with Lightning.

    Examples
    ========
    >>> import torch
    >>> ensemble = TorchMLPEnsemble(n_input=4, n_units=4, n_out=1, ensemble_size=2) 
    >>> ensemble(torch.randn(5, 4)).shape 
    torch.Size([5, 2])
    >>> ensemble.set_model(0)  # keep only the first sub-model
    >>> ensemble(torch.randn(2, 4)).shape 
    torch.Size([2, 1])
    >>> ensemble.set_model("all"); len(ensemble.model_ensemble) 
    2
    >>> list(ensemble.model_keys) == ["ensemble_module:0", "ensemble_module:1"] 
    True

    """

    def __init__(
        self, 
        ensemble_size: int = 1,
        learning_rate: float = .01,
        optimizer: Optimizer = Adam,
        reduce_lr_on_plateau: bool = True,
        reduce_lr_patience: int = 10,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.save_hyperparameters()
        self._init_ensemble(ensemble_size)
        self._init_lightning(
            optimizer=optimizer, 
            learning_rate=learning_rate, 
            model_attr='_model_ensemble',
            reduce_lr_on_plateau=reduce_lr_on_plateau,
            reduce_lr_patience=reduce_lr_patience,
        )

    def create_module(self) -> Module:
        return self.build_model()



"""For plotting splits."""

from typing import Iterable, Optional, Union

from datasets import Dataset
from carabiner import cast, print_err
from carabiner.mpl import add_legend, grid
import numpy as np
from schemist.converting import convert_string_representation
from schemist.features import calculate_feature
from schemist.tables import converter
import scipy
import umap

from .utils import dataset_len
from ..datasets import to_dataset 


def _check_columns(
    dataset: Dataset,
    columns: Optional[Iterable[str]] = None
):
    if columns is None:
        columns = []
    else:
        missing_from_data = [
            col for col in columns 
            if col not in dataset.column_names
        ]
        if len(missing_from_data) > 0:
            raise KeyError(f"Columns missing from dataset: {', '.join(missing_from_data)}")
    return columns


def plot_chemical_splits(
    ds: Dataset,
    structure_column: str = 'smiles',
    input_representation: str = 'smiles',
    split_columns: Optional[Iterable[str]] = None,
    descriptors: Optional[Union[str, Iterable[str]]] = None,
    sample_size: int = 20_000,
    additional_columns: Optional[Union[str, Iterable[str]]] = None,
    seed: int = 42,
    cache: str = "./cache",
    clean_smiles_column: str = "__clean_smiles__"
):

    """

    """
    
    if descriptors is None:
        descriptors = ['mwt', 'clogp', 'tpsa', 'max_charge', 'min_charge']
    else:
        descriptors = cast(descriptors, to=list)

    if split_columns is None:
        split_columns = ["split"]
    else:
        split_columns = cast(split_columns, to=list)
    if additional_columns is None:
        additional_columns = []
    else:
        additional_columns = cast(additional_columns, to=list)

    additional_columns = _check_columns(ds, additional_columns) + [col for col in descriptors if col in ds.column_names]
    descriptors = [col for col in descriptors if col not in ds.column_names]
    split_columns = _check_columns(ds, split_columns)

    cols_to_plot = sorted(set(descriptors + additional_columns + split_columns))

    print_err(
        f"Plotting UMAP embedding with maximum {sample_size} rows, coloring by", 
        ", ".join(cols_to_plot)
    )

    if dataset_len(ds) > sample_size:
        if isinstance(ds, Dataset):
            ds = ds.to_iterable_dataset(num_shards=min(1024, dataset_len(ds)))
        ds = ds.shuffle(seed=seed, buffer_size=1024).take(sample_size)
        ds = to_dataset(ds, nrows=sample_size, cache=cache)

    ds = ds.map(
        lambda x: {
            clean_smiles_column: convert_string_representation(
                strings=x[structure_column],
                input_representation=input_representation,
                output_representation="smiles",
            )
        },
    )

    df = (
        ds
        .to_pandas(batched=False)
        .assign(points=lambda x: np.nan * np.ones((x.shape[0], )))
    )
    errors, df = converter(
        df, 
        output_representation=descriptors,
    )
    fps, _ = calculate_feature(
        strings=df[clean_smiles_column],
        feature_type="fp",
        on_bits=True,
        return_dataframe=False,
    )
    fps = [sorted(map(int, fp.split(";"))) for fp in fps]
    fp_matrix = scipy.sparse.lil_matrix(
        (len(fps), 2048), 
        dtype=np.float32,
    )
    fp_matrix.rows = np.array(fps, dtype=object)
    fp_matrix.data = np.array(
        [np.ones(len(row), dtype=np.float32).tolist() for row in fps],
        dtype=object
    )
    fp_matrix = fp_matrix.tocsr()

    umapper = umap.UMAP(
        metric='jaccard', 
        min_dist=.1,
        random_state=seed, 
        low_memory=True,
    )
    umapper.fit(fp_matrix)
    fp_umapped = umapper.transform(fp_matrix)
    df = df.assign(
        umap1=fp_umapped[:,0], 
        umap2=fp_umapped[:,1],
    )

    fig, axes = grid(
        ncol=len(cols_to_plot) + 1, 
        aspect_ratio=1.2,
        sharex='row', 
        sharey='row', 
    )
    common_plot_kwargs = {
        "x": "umap1",
        "y": "umap2",
        "s": 1., 
    }
        
    for ax, col in zip(axes, ["points"] + cols_to_plot):
        ax.scatter(
            **common_plot_kwargs,
            data=df,
            c='lightgrey',
            zorder=-5,
        )
        if col in split_columns:
            unique_labels = sorted(df[col].unique())
            binary = (len(unique_labels) == 2 and (unique_labels[0] in (True, 0)))
            for i, label in enumerate(unique_labels):
                ax.scatter(
                    **(common_plot_kwargs | ({"s": 5.} if (label in (True, 1) and binary) else {})),
                    c="lightgrey" if (label in (False, 0) and binary) else f"C{i}",
                    data=df.query(f"{col} == @label"),
                    zorder=1,
                    label=label,
                )
            add_legend(ax)
        else:
            non_na_vals = df[col][~df[col].isna()]
            if all([
                np.all(non_na_vals >= 0.),
                len(non_na_vals) > 0,
                col not in ('min_charge', 'max_charge', 'points'),
                non_na_vals.min() != non_na_vals.max(),
            ]):
                log_colors = True
            else:
                log_colors = False
            sc = ax.scatter(
                **common_plot_kwargs,
                c=col,
                data=df,
                cmap='magma', 
                norm="log" if log_colors else None,
                zorder=1,
            )
            if col != 'points':
                try:
                    fig.colorbar(sc, ax=ax)
                except ValueError as e:
                    print_err(e)
        ax.set(title=col)
        ax.set_axis_off()
    
    return (fig, axes), df





from typing import Union
from datasets import Dataset, IterableDataset
from .faiss import faiss_split
from .scaffold import scaffold_split

SPLIT_METHODS = {
    "scaffold": scaffold_split,
    "faiss": faiss_split,
}

def split_dataset(
    ds: Union[Dataset, IterableDataset],
    method: str,
    *args, **kwargs
):  
    try:
        return SPLIT_METHODS[method](ds, *args, **kwargs)
    except KeyError:
        raise KeyError(f"No method '{method}' for splitting")
